from enum import Enum


class ConcentrationTypes(Enum):
    MOLAR = "mole / l"
    MILLIMOLAR = "mmole / l"
    MICROMOLAR = "umole / l"
    NANAMOLAR = "nmole / l"
    GRAMLITER = "g / l"
    MILLIGRAMLITER = "mg / l"
    MICROGRAMLITER = "ug / l"
    NANGRAMLITER = "ng / l"
